#version 330 core

const int MAX_POINT_LIGHTS = 4; 

in vec2 fragmentUV;
in vec3 fragmentNormal;
in vec3 worldPos;
in vec4 lightSpacePos;

layout(location=0) out vec4 fragColor;

struct BaseLight                                                                    
{                                                                                   
    vec3 Color;                                                                     
    float AmbientIntensity;                                                         
    float DiffuseIntensity;                                                         
};                                                                                  
                                                                                    
struct DirectionalLight                                                             
{                                                                                   
    BaseLight Base;                                                                 
    vec3 Direction;    
	mat4 Rotation;                                                             
};                                                                                  
                                                                                    
struct Attenuation                                                                  
{                                                                                   
    float Constant;                                                                 
    float Linear;                                                                   
    float Exp;                                                                      
};                                                                                  
                                                                                    
struct PointLight                                                                           
{                                                                                           
    BaseLight Base;                                                                         
    vec3 Position;                                                                          
    Attenuation Atten;                                                                      
};

struct Material
{
	vec4 ambient;
	vec4 texture_diffuse1;
	vec4 texture_specular1;
	float shininess;
};

uniform sampler2D texture_diffuse1;
uniform sampler2DShadow shadowMap;
uniform vec2 mapSize;
uniform DirectionalLight directionalLight;
uniform vec3 eyeWorldPos;
uniform float matSpecularIntensity;
uniform int numPointLights; 
uniform PointLight pointLights[MAX_POINT_LIGHTS];  
uniform Material material;
//uniform vec4 mEmission;

uniform samplerCubeShadow cubemap;
uniform vec2 nearFar;

#define EPSILON 0.009

vec3 gridSamplingDisk[20] = vec3[]
(
   vec3(1, 1, 1), vec3(1, -1, 1), vec3(-1, -1, 1), vec3(-1, 1, 1), 
   vec3(1, 1, -1), vec3(1, -1, -1), vec3(-1, -1, -1), vec3(-1, 1, -1),
   vec3(1, 1, 0), vec3(1, -1, 0), vec3(-1, -1, 0), vec3(-1, 1, 0),
   vec3(1, 0, 1), vec3(-1, 0, 1), vec3(1, 0, -1), vec3(-1, 0, -1),
   vec3(0, 1, 1), vec3(0, -1, 1), vec3(0, -1, -1), vec3(0, 1, -1)
);


float CalcOmindirectionalShadowFactor(PointLight pointLight)
{	
	// vector from surface point to light pos
	vec3 cubemapLookup = worldPos - pointLight.Position;
	float distanceToLight = length(cubemapLookup);
	float currentDistanceToLight = (distanceToLight - nearFar.x)/(nearFar.y - nearFar.x);
   currentDistanceToLight = clamp(currentDistanceToLight, 0, 1);
    // radius of PCF depending on distance from the light source
   float diskRadius = (1.0 + (1.0 - currentDistanceToLight) * 3) / 1024;

   	float Factor = 0.0;

   // evaluate each sampling direction
   for(int i=0; i<20; i++)
   {
		Factor += texture(cubemap, vec4(normalize(cubemapLookup) + (gridSamplingDisk[i] * diskRadius), currentDistanceToLight-0.003));
   }
	return 0.4 + (Factor/20.0);
}


float CalcShadowFactor(vec4 LightSpacePos, vec3 normal, vec3 lightDirection)
{
	vec3 ProjCoords = LightSpacePos.xyz / LightSpacePos.w;
    vec2 UVCoords;
    UVCoords.x = 0.5 * ProjCoords.x + 0.5;
    UVCoords.y = 0.5 * ProjCoords.y + 0.5;
    float z = 0.5 * ProjCoords.z + 0.5;

    float xOffset = 1.0/mapSize.x;
    float yOffset = 1.0/mapSize.y;

    float Factor = 0.0;
    for (int y = -1 ; y <= 1 ; y++) {
        for (int x = -1 ; x <= 1 ; x++) {
            vec2 Offsets = vec2(x * xOffset, y * yOffset);
            vec3 UVC = vec3(UVCoords + Offsets, z-0.00001);
            Factor += texture(shadowMap, UVC);
        }
    }
    return ((Factor/18.0));
}

vec4 CalcLightInternal(BaseLight light, vec3 lightDirection, vec3 normal, float shadowFactor)                   
{                                                                                           
    vec4 ambientColor = material.ambient * vec4(light.Color, 1.0f) * light.AmbientIntensity;                   
    float diffuseFactor = dot(normal, -lightDirection);                                     
                                                                                            
    vec4 diffuseColor  = vec4(0, 0, 0, 0);                                                  
    vec4 specularColor = vec4(0, 0, 0, 0);                                                  
                                                                                            
    if (diffuseFactor > 0) {                                                                
        diffuseColor = material.texture_diffuse1 * vec4(light.Color, 1.0f) * light.DiffuseIntensity * diffuseFactor;    
                                                                                            
        vec3 vertexToEye = normalize(eyeWorldPos - worldPos);                             
        vec3 lightReflect = normalize(reflect(lightDirection, normal));                     
        float specularFactor = dot(vertexToEye, lightReflect);                              
        specularFactor = pow(specularFactor, material.shininess);                               
        if (specularFactor > 0) {                                                           
            specularColor =material.texture_specular1 * vec4(light.Color, 1.0f) *                                       
                            matSpecularIntensity * specularFactor;                         
        }                                                                                   
    }                                                                                       
                                                                                            
    return (ambientColor + (shadowFactor *(diffuseColor + specularColor)));                                   
}     

vec4 CalcDirectionalLight(vec3 normal)                                                      
{                                                     
	vec3 dir =  normalize(directionalLight.Rotation * vec4(directionalLight.Direction,1)).xyz;                                     
    return CalcLightInternal(directionalLight.Base, dir, normal,  CalcShadowFactor(lightSpacePos, normal, dir)); 
}             

vec4 CalcPointLight(PointLight l, vec3 normal, vec4 LightSpacePos)                                                 
{                                                                                           
    vec3 lightDirection = worldPos - l.Position;                         
    float distance = length(lightDirection);                                                
    lightDirection = normalize(lightDirection);                                             
	float OmnidirShadowFactor = CalcOmindirectionalShadowFactor(l);
						                                                                    
    vec4 color = CalcLightInternal(l.Base, lightDirection, normal, OmnidirShadowFactor);       
    float attenuation =  l.Atten.Constant +                               
                         l.Atten.Linear * distance +                      
                         l.Atten.Exp * distance * distance;               
                                                                                            
    return color / attenuation;                                                             
}              

void main()
{
	vec3 normal = normalize(fragmentNormal);                                                       
    vec4 totalLight = CalcDirectionalLight(normal);                                         
                                                                                            
    for (int i = 0 ; i < numPointLights ; i++) {                                           
        totalLight += CalcPointLight(pointLights[i], normal, lightSpacePos);                                            
    }                                                                                       	      
	
    fragColor = texture2D(texture_diffuse1, fragmentUV.xy) * totalLight;
}